#ifndef INFO_HEADER
#define INFO_HEADER

#include "Absyn.h"
#include <memory>
#include <set>
using namespace std;

#include "InfoList.h"

class VarInfo {
public:
	VarInfo(PIdent* ident, Type* type = NULL) : ident(ident), name(ident->string_), type(type), lineLocation(ident->integer_) {}
	VarInfo(Ar* arg) : VarInfo(arg->pident_, arg->type_) {}

	PIdent *ident;
	string name;
	Type *type;
	int lineLocation;

	bool operator<(const VarInfo &other) const {
		return name != other.name ? name < other.name : type < other.type;
	}
};


class FunctionInfo : public VarInfo {
public:
	Block * block;
	vector<VarInfoPtr> arguments;
	weak_ptr<ClassInfo> klass;
	weak_ptr<Binding> binding;
	FunctionInfo(FuncDef *expr, ClassInfoPtr klass = nullptr);
	// FunctionInfo(PIdent *ident, ClassInfoPtr klass = nullptr) : VarInfo(ident), block(NULL), klass(klass) {};
};



class Binding {
public:
	InfoList<VarInfo> variables;
	BindingPtr parent;
	Binding(BindingPtr parent);

	BindingPtr getParent() const { return parent; };
};

class ClassInfo : public VarInfo, public Binding
{
public:
	InfoList<FunctionInfo> functions;

	ClassInfo(PIdent *ident, ClassInfoPtr parent = nullptr);

	ClassInfoPtr getParent() const { return static_pointer_cast<ClassInfo>(parent); };
};


class Scope {
public:
	InfoList<ClassInfo> classes;
	InfoList<FunctionInfo> functions;
	BindingPtr currentBinding;
	ClassInfoPtr currentClass;
	FunctionInfoPtr currentFunction;
	Scope();
};

#endif