#ifndef INFOLIST_HEADER
#define INFOLIST_HEADER

#include "Info.h"
#include "ParseError.h"
#include "TypeDefs.h"

template<typename T>
struct Compare
{
	using is_transparent = void;
	bool operator() (const T& a, const T& b) const { return a->name < b->name; }
	bool operator() (const T& a, const string &b) const { return a->name < b; }
	bool operator() (const string &a, const T& b) const { return a < b->name; }
	// bool operator() (const T& a, const PIdent *b) const { return a->name < b->string_; }
	// bool operator() (const PIdent *a, const T& b) const { return a->string_ < b->name; }
};


template<typename T,
	typename = typename std::enable_if<std::is_base_of<VarInfo, T>::value>::type,
	typename TPtr = shared_ptr<T>>
class InfoList : public set<TPtr, Compare<TPtr>> {
public:
	InfoList<T> *parent;
	InfoList(InfoList<T> *parent = nullptr) : parent(parent) {};
	TPtr operator[](PIdent *ident) const {
		TPtr ret = local(ident->string_);
		if (ret) return ret;
		if (parent) return (*parent)[ident];
		throw UndefinedError(ident);
	}

	TPtr local(string name) const {
		const auto p = this->find(name);
		return p != this->end() ? *p : nullptr;
	}

	void add(TPtr obj) {
		auto ret = this->emplace(obj);
		if (!ret.second) {
			throw RedefinedError(obj->ident, *ret.first);
		}
	}

	InfoList<T>& operator<<(TPtr obj) {
		add(obj);
		return *this;
	}
};

#endif
