#include "TypeCheck.h"
#include "ParseError.h"
#include <stdexcept>

void TypeCheck::visitClassDef(ClassDef *t) {
	const string name = t->getName()->string_;
	ClassInfoPtr parent = NULL;
	if (t->getParent()) {
		parent = scope.classes[t->getParent()];
		if (!parent) {
			throw UndefinedError(t->getParent());
		}
	}
	auto c = make_shared<ClassInfo>(t->getName(), parent);
	scope.classes << c;

	scope.currentClass = c;
	t->getBlock()->accept(this);
	scope.currentClass = NULL;
}

void TypeCheck::visitPIdent(PIdent *p_ident)
{
	visitString(p_ident->string_);
	visitInteger(p_ident->integer_);
}

void TypeCheck::visitProg(Prog *prog)
{
	/* Code For Prog Goes Here */

	prog->listtopdef_->accept(this);

}

void TypeCheck::visitFnDef(FnDef *fn_def)
{
	/* Code For FnDef Goes Here */

	fn_def->fundef_->accept(this);

}

void TypeCheck::visitClDef(ClDef *cl_def)
{
	/* Code For ClDef Goes Here */

	cl_def->classdef_->accept(this);

}

void TypeCheck::visitFuncDef(FuncDef *def)
{
	FunctionInfoPtr f = make_shared<FunctionInfo>(def, scope.currentClass);

	// FunctionInfo tworzy argumenty

	auto &targetVector = scope.currentClass ? scope.currentClass->functions : scope.functions;
	targetVector << f;
}

void TypeCheck::visitClassDefN(ClassDefN *class_def_n)
{
	visitClassDef(class_def_n);
}

void TypeCheck::visitClassDefE(ClassDefE *class_def_e)
{
	visitClassDef(class_def_e);
}

void TypeCheck::visitClassBl(ClassBl *class_bl)
{
	/* Code For ClassBl Goes Here */

	class_bl->listclassblockdef_->accept(this);

}

void TypeCheck::visitClassMthd(ClassMthd *class_mthd)
{
	/* Code For ClassMthd Goes Here */

	class_mthd->fundef_->accept(this);

}

void TypeCheck::visitClassFld(ClassFld *class_fld)
{
	/* Code For ClassFld Goes Here */

	class_fld->type_->accept(this);
	class_fld->listitem_->accept(this);

}

void TypeCheck::visitBlk(Blk *blk)
{
	BindingPtr binding = blk->binding.lock();
	if (!binding) {
		binding = make_shared<Binding>(scope.currentBinding);
		blk->binding = binding;
	}

	scope.currentBinding = binding;

	blk->liststmt_->accept(this);

	scope.currentBinding = binding->getParent();
}

void TypeCheck::visitBStmt(BStmt *b_stmt)
{
	b_stmt->block_->accept(this);
}

void TypeCheck::visitDecl(Decl *decl)
{
	Type* type = decl->type_;

	for (Item * el : *decl->listitem_) {
		el->expr_->accept(this);

		VarInfoPtr var = make_shared<VarInfo>(el->pident_, type);
		scope.currentBinding.variables << var;
	}
}

void TypeCheck::visitAss(Ass *ass)
{
	auto var = scope.currentBinding.variables[]
	ass->expr_1->accept(this);
	ass->expr_2->accept(this);

}

void TypeCheck::visitTableAss(TableAss *table_ass)
{
	/* Code For TableAss Goes Here */

	table_ass->pident_->accept(this);
	table_ass->expr_1->accept(this);
	table_ass->expr_2->accept(this);

}

void TypeCheck::visitTableIncr(TableIncr *table_incr)
{
	/* Code For TableIncr Goes Here */

	table_incr->pident_->accept(this);
	table_incr->expr_->accept(this);

}

void TypeCheck::visitTableDecr(TableDecr *table_decr)
{
	/* Code For TableDecr Goes Here */

	table_decr->pident_->accept(this);
	table_decr->expr_->accept(this);

}

void TypeCheck::visitIncr(Incr *incr)
{
	/* Code For Incr Goes Here */

	incr->pident_->accept(this);

}

void TypeCheck::visitDecr(Decr *decr)
{
	/* Code For Decr Goes Here */

	decr->pident_->accept(this);

}

void TypeCheck::visitRet(Ret *ret)
{
	/* Code For Ret Goes Here */

	ret->expr_->accept(this);

}

void TypeCheck::visitVRet(VRet *v_ret)
{
	/* Code For VRet Goes Here */


}

void TypeCheck::visitCond(Cond *cond)
{
	/* Code For Cond Goes Here */

	cond->expr_->accept(this);
	cond->stmt_->accept(this);

}

void TypeCheck::visitCondElse(CondElse *cond_else)
{
	/* Code For CondElse Goes Here */

	cond_else->expr_->accept(this);
	cond_else->stmt_1->accept(this);
	cond_else->stmt_2->accept(this);

}

void TypeCheck::visitWhile(While *while_)
{
	/* Code For While Goes Here */

	while_->expr_->accept(this);
	while_->stmt_->accept(this);

}

void TypeCheck::visitSExp(SExp *s_exp)
{
	/* Code For SExp Goes Here */

	s_exp->expr_->accept(this);

}

void TypeCheck::visitForEach(ForEach *for_each)
{
	/* Code For ForEach Goes Here */

	for_each->type_->accept(this);
	for_each->pident_->accept(this);
	for_each->expr_->accept(this);
	for_each->stmt_->accept(this);

}

void TypeCheck::visitInt(Int *int_)
{
	/* Code For Int Goes Here */


}

void TypeCheck::visitStr(Str *str)
{
	/* Code For Str Goes Here */


}

void TypeCheck::visitBool(Bool *bool_)
{
	/* Code For Bool Goes Here */


}

void TypeCheck::visitVoid(Void *void_)
{
	/* Code For Void Goes Here */


}

void TypeCheck::visitArray(Array *array)
{
	/* Code For Array Goes Here */

	array->type_->accept(this);

}

void TypeCheck::visitClassT(ClassT *class_t)
{
	/* Code For ClassT Goes Here */

	class_t->pident_->accept(this);

}

void TypeCheck::visitFun(Fun *fun)
{
	/* Code For Fun Goes Here */

	fun->type_->accept(this);
	fun->listtype_->accept(this);

}

void TypeCheck::visitEVar(EVar *e_var)
{
	/* Code For EVar Goes Here */

	e_var->pident_->accept(this);

}

void TypeCheck::visitELitInt(ELitInt *e_lit_int)
{
	visitInteger(e_lit_int->integer_);
	lastType = make_shared<Int>();
}

void TypeCheck::visitELitTrue(ELitTrue *e_lit_true)
{
	lastType = make_shared<Bool>();
}

void TypeCheck::visitELitFalse(ELitFalse *e_lit_false)
{
	lastType = make_shared<Bool>();
}

void TypeCheck::visitEApp(EApp *e_app)
{
	/* Code For EApp Goes Here */

	e_app->pident_->accept(this);
	e_app->listexpr_->accept(this);

}

void TypeCheck::visitEString(EString *e)
{
	/* Code For EString Goes Here */

	visitString(e->string_);
	auto a = lastType;
	Str expect;
	if (*a != expect) {
		throw InvalidTypeError(e->lineno, expect, {a}, e);
	}
	lastType = make_shared<Str>();
}

void TypeCheck::visitENewArray(ENewArray *e)
{
	/* Code For ENewArray Goes Here */

	e->type_->accept(this);
	e->expr_->accept(this);
	auto a = lastType;
	Array expect(e->type_);
	if (*a != expect) {
		throw InvalidTypeError(e->lineno, expect, {a}, e);
	}
	lastType = make_shared<Array>(e->type_);
}

void TypeCheck::visitENewClass(ENewClass *e_new_class)
{
	e_new_class->pident_->accept(this);
	lastType = make_shared<ClassT>(e_new_class->pident_);
}

void TypeCheck::visitEClsMmbr(EClsMmbr *e_cls_mmbr)
{
	/* Code For EClsMmbr Goes Here */

	e_cls_mmbr->expr_->accept(this);
	e_cls_mmbr->pident_->accept(this);

}

void TypeCheck::visitEClsMthd(EClsMthd *e_cls_mthd)
{
	/* Code For EClsMthd Goes Here */

	e_cls_mthd->expr_->accept(this);
	e_cls_mthd->pident_->accept(this);
	e_cls_mthd->listexpr_->accept(this);

}

void TypeCheck::visitNull(Null *e)
{
	PIdent *ident = new PIdent("null", e->lineno);
	lastType = make_shared<ClassT>(ident); // TODO
}

void TypeCheck::visitEIndexAcc(EIndexAcc *e_index_acc)
{
	/* Code For EIndexAcc Goes Here */

	e_index_acc->pident_->accept(this);
	e_index_acc->expr_->accept(this);
	throw UndefinedError(e_index_acc->pident_); // TODO
}

void TypeCheck::visitECast(ECast *e_cast)
{
	/* Code For ECast Goes Here */

	e_cast->pident_->accept(this);
	e_cast->expr_->accept(this);
	lastType = make_shared<ClassT>(e_cast->pident_);
}

void TypeCheck::visitNeg(Neg *e)
{
	e->expr_->accept(this);
	auto a = lastType;
	Int expect;
	if (*a != expect) {
		throw InvalidTypeError(e->lineno, expect, {a}, e);
	}
	lastType = make_shared<Int>();
}

void TypeCheck::visitNot(Not *e)
{
	e->expr_->accept(this);
	auto a = lastType;
	Bool expect;
	if (*a != expect) {
		throw InvalidTypeError(e->lineno, expect, {a}, e);
	}
	lastType = make_shared<Bool>();
}

void TypeCheck::visitEMul(EMul *e)
{
	e->expr_1->accept(this);
	auto a = lastType;
	e->expr_2->accept(this);
	auto b = lastType;
	Int expect;
	if (*a != expect || *a != *b) {
		throw InvalidTypeError(e->lineno, expect, {a, b}, e);
	}
	lastType = make_shared<Int>();
}

void TypeCheck::visitEAdd(EAdd *e)
{
	e->expr_1->accept(this);
	auto a = lastType;
	e->expr_2->accept(this);
	auto b = lastType;
	Int expect;
	if (*a != expect || *a != *b) {
		throw InvalidTypeError(e->lineno, expect, {a, b}, e);
	}
	lastType = make_shared<Int>();
}

void TypeCheck::visitERel(ERel *e)
{
	e->expr_1->accept(this);
	auto a = lastType;
	e->expr_2->accept(this);
	auto b = lastType;
	Int expect;
	if (*a != expect || *a != *b) {
		throw InvalidTypeError(e->lineno, expect, {a, b}, e);
	}
	lastType = make_shared<Bool>();
}

void TypeCheck::visitEAnd(EAnd *e)
{
	e->expr_1->accept(this);
	auto a = lastType;
	e->expr_2->accept(this);
	auto b = lastType;
	Bool expect;
	if (*a != expect || *a != *b) {
		throw InvalidTypeError(e->lineno, expect, {a, b}, e);
	}
	lastType = make_shared<Bool>();
}

void TypeCheck::visitEOr(EOr *e)
{
	e->expr_1->accept(this);
	auto a = lastType;
	e->expr_2->accept(this);
	auto b = lastType;
	Bool expect;
	if (*a != expect || *a != *b) {
		throw InvalidTypeError(e->lineno, expect, {a, b}, e);
	}
	lastType = make_shared<Bool>();
}

void TypeCheck::visitListTopDef(ListTopDef *list_top_def)
{
	for (auto &i : *list_top_def)
	{
		i->accept(this);
	}
}

void TypeCheck::visitListArg(ListArg *list_arg)
{
	for (ListArg::iterator i = list_arg->begin() ; i != list_arg->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListClassBlockDef(ListClassBlockDef *list_class_block_def)
{
	for (ListClassBlockDef::iterator i = list_class_block_def->begin() ; i != list_class_block_def->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListStmt(ListStmt *list_stmt)
{
	for (ListStmt::iterator i = list_stmt->begin() ; i != list_stmt->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListItem(ListItem *list_item)
{
	for (ListItem::iterator i = list_item->begin() ; i != list_item->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListType(ListType *list_type)
{
	for (ListType::iterator i = list_type->begin() ; i != list_type->end() ; ++i)
	{
		(*i)->accept(this);
	}
}

void TypeCheck::visitListExpr(ListExpr *list_expr)
{
	for (ListExpr::iterator i = list_expr->begin() ; i != list_expr->end() ; ++i)
	{
		(*i)->accept(this);
	}
}


void TypeCheck::visitInteger(Integer x)
{
	/* Code for Integer Goes Here */
}

void TypeCheck::visitString(String x)
{
	/* Code for String Goes Here */
}

TypeCheck::TypeCheck()
	: state(initialized)
{
}

void TypeCheck::check(Visitable *v)
{
	if (state != State::initialized) {
		throw std::runtime_error("already initialized");
	}
	state = State::checkType;
	v->accept(this);

	for (auto c : scope.classes) {
		scope.currentClass = c;

		for (auto f : c->functions) {
			checkFunction(f);
		}

		scope.currentClass = nullptr;
	}
}

void TypeCheck::checkFunction(FunctionInfoPtr f)
{
	scope.currentFunction = f;

	BindingPtr binding = make_shared<Binding>(nullptr);
	f->binding = binding;

	for (auto arg : f->arguments) {
		binding->variables << arg;
	}

	f->block->accept(this);

	scope.currentFunction = nullptr;
}
