#ifndef TYPECHECK_HEADER
#define TYPECHECK_HEADER
/* You might want to change the above name. */

#include "Absyn.h"
#include "Info.h"
#include <memory>


class TypeCheck : public Visitor
{
	enum State {
		initialized,
		buildInfo,
		checkType
	} state;

	Scope scope;
	shared_ptr<Type> lastType;

	void checkFunction(FunctionInfoPtr f);
public:
	TypeCheck();
	void check(Visitable *v);
protected:
	void visitClassDef(ClassDef *p);
	void visitProg(Prog *p);
	void visitFnDef(FnDef *p);
	void visitClDef(ClDef *p);
	void visitFuncDef(FuncDef *p);
	void visitClassDefN(ClassDefN *p);
	void visitClassDefE(ClassDefE *p);
	void visitClassBl(ClassBl *p);
	void visitClassMthd(ClassMthd *p);
	void visitClassFld(ClassFld *p);
	void visitBlk(Blk *p);
	void visitBStmt(BStmt *p);
	void visitDecl(Decl *p);
	void visitAss(Ass *p);
	void visitTableAss(TableAss *p);
	void visitTableIncr(TableIncr *p);
	void visitTableDecr(TableDecr *p);
	void visitIncr(Incr *p);
	void visitDecr(Decr *p);
	void visitRet(Ret *p);
	void visitVRet(VRet *p);
	void visitCond(Cond *p);
	void visitCondElse(CondElse *p);
	void visitWhile(While *p);
	void visitSExp(SExp *p);
	void visitForEach(ForEach *p);
	void visitInt(Int *p);
	void visitStr(Str *p);
	void visitBool(Bool *p);
	void visitVoid(Void *p);
	void visitArray(Array *p);
	void visitClassT(ClassT *p);
	void visitFun(Fun *p);
	void visitEVar(EVar *p);
	void visitELitInt(ELitInt *p);
	void visitELitTrue(ELitTrue *p);
	void visitELitFalse(ELitFalse *p);
	void visitEApp(EApp *p);
	void visitEString(EString *p);
	void visitENewArray(ENewArray *p);
	void visitENewClass(ENewClass *p);
	void visitEClsMmbr(EClsMmbr *p);
	void visitEClsMthd(EClsMthd *p);
	void visitNull(Null *p);
	void visitEIndexAcc(EIndexAcc *p);
	void visitECast(ECast *p);
	void visitNeg(Neg *p);
	void visitNot(Not *p);
	void visitEMul(EMul *p);
	void visitEAdd(EAdd *p);
	void visitERel(ERel *p);
	void visitEAnd(EAnd *p);
	void visitEOr(EOr *p);
	void visitListTopDef(ListTopDef *p);
	void visitListArg(ListArg *p);
	void visitListClassBlockDef(ListClassBlockDef *p);
	void visitListStmt(ListStmt *p);
	void visitListItem(ListItem *p);
	void visitListType(ListType *p);
	void visitListExpr(ListExpr *p);
	void visitPIdent(PIdent *p);

	void visitInteger(Integer x);
	void visitChar(Char x) {};
	void visitString(String x);

// abstract

	// topdefs

	void visitProgram(Program *t) {} //abstract class
	void visitTopDef(TopDef *t) {} //abstract class
	void visitFunDef(FunDef *t) {} //abstract class
	void visitBlock(Block *t) {};
	void visitClassBlock(ClassBlock *p) {};
	void visitClassBlockDef(ClassBlockDef *p) {};
	void visitItem(Item *p) {};
	void visitArg(Arg *p) {}
	void visitStmt(Stmt *p) {};
	void visitType(Type *p) {};
	void visitExpr(Expr *p) {};
	void visitAddOp(AddOp *p) {};
	void visitMulOp(MulOp *p) {};
	void visitRelOp(RelOp *p) {};

	// stmts
	void visitEmpty(Empty *p) {};
	void visitNoInit(NoInit *p) {};
	void visitInit(Init *p) {};
	void visitAr(Ar *p) {};

	// exprs
	void visitPlus(Plus *p) {};
	void visitMinus(Minus *p) {};
	void visitTimes(Times *p) {};
	void visitDiv(Div *p) {};
	void visitMod(Mod *p) {};
	void visitLTH(LTH *p) {};
	void visitLE(LE *p) {};
	void visitGTH(GTH *p) {};
	void visitGE(GE *p) {};
	void visitEQU(EQU *p) {};
	void visitNE(NE *p) {};

	// terminals
	void visitDouble(Double x) {};
	void visitIdent(Ident x) {};
};

#endif
